import os
from .decorators import auth_required,root_auth_required
from .validator import validate_token
from .exceptions import JWTValidationError
from .config import init_sso_config
from .permissions import Permission, BasePermission



import time

class Cache:
    def __init__(self, ttl_seconds=60*60):
        self.store = {}
        self.ttl = ttl_seconds

    def get(self, key):
        entry = self.store.get(key)
        if entry and entry['expiry'] > time.time():
            return entry['data']
        self.store.pop(key, None)
        return None

    def set(self, key, data):
        self.store[key] = {
            'data': data,
            'expiry': time.time() + self.ttl
        }

    def clear(self):
        self.store={}



class BasePermission:
    """
    Interface for any permission checker.
    """
    def __init__(self, claims: dict, permission: str = None, value=None, **kwargs):
        self.user = claims
        self.permission = permission
        self.value = value
        self.extras = kwargs

    def has_access(self) -> bool:
        raise NotImplementedError
    

class Permission(BasePermission):
    """
    Default permission checker for claims stored in nested JSON.
    """

    def __init__(self, claims: dict, permission: str = None, value=None):
        super().__init__(claims, permission, value)

    def has_access(self) -> bool:
        if not self.permission:
            return True 

        keys = self.permission.split(".")
        node = self.user.get('permissions', {})

        try:
            for key in keys:
                if isinstance(node, dict):
                    node = node.get(key)
                else:
                    return False
        except Exception:
            return False

        if node == "*":
            return True 
    
        if node is None:
            return False

        if isinstance(self.value, list) and  not isinstance(node, (dict,list)):
                    return node in self.value
        
        if isinstance(node, list):
            if self.value is not None:
                if isinstance(self.value, list):
                    return any(v in node for v in self.value)
                
                return self.value in node
            return False

        # Primitive value comparison or truthy check if no value provided
        return node == self.value if self.value is not None else bool(node)

        


# decorators.py

from functools import wraps
from more_sso.validator import validate_jwt
from more_sso.exceptions import AccessDeniedError, JWTValidationError
from typing import TypeVar
import json
from more_sso.permissions import Permission

RESPONSE_HEADERS = {
    "Content-Type": "application/json",
    "Access-Control-Allow-Origin": "*",
    "Access-Control-Allow-Headers":"*.more.in",
    "Access-Control-Allow-Methods": "GET, POST, OPTIONS",
    }

def json_response(status_code: int, detail: str="success", data: dict = {}):
    return {
        "statusCode": status_code,
        "headers":RESPONSE_HEADERS,
        "body": json.dumps({
            "detail": detail,
            "data": data or {}
        })
    }


def auth_required(permission_class=Permission, permission: str = None, value=None,**kwargs):
    def decorator(func):
        @wraps(func)
        def wrapper(event: dict , *args, **kwargs):
            token = event.get("headers", {}).get("Authorization", "") or  event.get("headers", {}).get("authorization", "")
            if not token:
                return json_response( 401, detail= "Unauthorized: Missing or invalid Authorization header")
            try:
                claims = validate_jwt(token)
                event['requestContext']['user'] = claims
                permission_obj = permission_class(claims, permission, value, **kwargs)

                if not permission_obj.has_access():
                    raise AccessDeniedError

                return func(event, *args, **kwargs)
            except JWTValidationError as e:
                return json_response( 401, detail= str(e) )
        return wrapper
    
    return decorator

def root_auth_required(func):
    @wraps(func)
    def wrapper(event, context):
        token = event.get("headers", {}).get("Authorization", "") or  event.get("headers", {}).get("authorization", "")
        if not token :
            return json_response( 401, detail= "Unauthorized: Missing or invalid Authorization header")
        try:
            user = validate_jwt(token)
            if not "requestContext" in event:
                event['requestContext'] = {}
            event['requestContext']['user'] = user
            return func(event, context)
        except JWTValidationError as e:
            return json_response( 401, detail=str(e))
    return wrapper


import json
import os
import boto3
import base64
class ConfigMissingError(Exception):
    """Raised when required SSO config is missing."""
    pass

REQUIRED_SSO_KEYS = ["public_key_uri"]
ADDITIONAL_KEYS = ["audience"]

def get_pem(KeyId) -> str:
    kms = boto3.client("kms",'ap-south-1')
    resp = kms.get_public_key(KeyId=KeyId)
    b:bytes = resp["PublicKey"]
    body = base64.encodebytes(b).decode("ascii").replace("\n", "")
    lines = [body[i:i+64] for i in range(0, len(body), 64)]
    return "-----BEGIN PUBLIC KEY-----\n" + "\n".join(lines) + "\n-----END PUBLIC KEY-----\n"

_config = None

def _validate_config(config: dict):
    missing = [key.upper() for key in REQUIRED_SSO_KEYS if not config.get(key)]
    if missing:
        raise ConfigMissingError(f"Missing SSO config: {', '.join(missing)} \n please set them in environment or pass them as parameters to init_sso_config( ) ")

def init_sso_config(public_key_uri=None,audience=None):
    """
    Initialize config from parameters or environment variables.
    Supports any keys. Required keys are checked dynamically.
    """
    global _config
    _config = {
        "public_key_uri": public_key_uri,
        "audience": audience
    }
    _validate_config(_config)


def get_sso_config():
    global _config
    if _config is not None:
        return _config

    # Lazy load from env if not already initialized
    config = {
        key: os.getenv(f"{key.upper()}", "")
        for key in REQUIRED_SSO_KEYS
    }

    for key in ADDITIONAL_KEYS:
        config[key] = os.getenv(f"{key.upper()}", "")

    _validate_config(config)
    _config = config
    return config




import json
import jwt
from more_sso.cache import Cache
from more_sso.config import get_sso_config,get_pem
from more_sso.exceptions import JWTValidationError
from functools import partial
_public_key_cache = Cache(ttl_seconds=60*60)

def get_public_key() -> str:
    cached_key = _public_key_cache.get('PUBLIC_KEY')
    if cached_key:
        return cached_key,_public_key_cache.get('AUDIENCE') 

    cfg = get_sso_config()
    public_key = get_pem(cfg['public_key_uri'])
    _public_key_cache.set('PUBLIC_KEY', public_key)
    _public_key_cache.set('AUDIENCE', cfg.get('audience'))  
    return public_key,cfg.get('audience')

def validate_jwt(token: str, options: dict = {}) -> dict:
    public_key,audience = get_public_key()
    decode_fn = jwt.decode
    if audience and options.get("verify_aud", True):
         decode_fn = partial(jwt.decode,audience=audience)
    try:
        if token.startswith("Bearer "):
            token = token.split("Bearer ")[1].strip()
        payload = decode_fn(
            token,
            token_type='access',
            key=public_key,
            algorithms=["RS256"],
            options=options
        )
        return payload
    except Exception as e:
        raise JWTValidationError(f"JWT validation failed: {str(e)}")

def validate_token(token, options:dict={}) -> dict:
    user = validate_jwt(token, options)
    return user